/*
 * High-level NDS client library
 *
 * Copyright (C) 2012  Leo Singer <leo.singer@ligo.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#if HAVE_CONFIG_H
#include "daq_config.h"
#endif /* HAVE_CONFIG_H */

#include "nds.h"

#include <daqc.h>
#include <daqc_response.h>
#include <channel.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#if HAVE_FNMATCH_H
#include <fnmatch.h>
#endif /* HAVE_FNMATCH_H */

#include "nds_os.h"

#ifndef FNM_NOMATCH
#define FNMATCH_STATIC 1
#define FNM_NOMATCH 1

static int fnmatch( const char* Pattern,
		    const char* Name,
		    int Flags );
#endif /* FNM_NOMATCH */


struct _nds_connection
{
	daq_t daq;
	char *host;
	int port;
};


static nds_channel_type _nds_chantype_to_channel_type(chantype_t c)
{
	nds_channel_type channel_type;

	switch (c)
	{
		case cOnline:
			channel_type = NDS_CHANNEL_TYPE_ONLINE;
			break;
		case cRaw:
			channel_type = NDS_CHANNEL_TYPE_RAW;
			break;
		case cRDS:
			channel_type = NDS_CHANNEL_TYPE_RDS;
			break;
		case cSTrend:
			channel_type = NDS_CHANNEL_TYPE_STREND;
			break;
		case cMTrend:
			channel_type = NDS_CHANNEL_TYPE_MTREND;
			break;
		case cTestPoint:
			channel_type = NDS_CHANNEL_TYPE_TEST_POINT;
			break;
		case cStatic:
			channel_type = NDS_CHANNEL_TYPE_STATIC;
			break;
		default:
			channel_type = 0;
	}

	return channel_type;
}


static nds_data_type _nds_data_to_data_type(daq_data_t c)
{
	nds_data_type data_type;

	switch (c)
	{
		case _16bit_integer:
			data_type = NDS_DATA_TYPE_INT16;
			break;
		case _32bit_integer:
			data_type = NDS_DATA_TYPE_INT32;
			break;
		case _64bit_integer:
			data_type = NDS_DATA_TYPE_INT64;
			break;
		case _32bit_float:
			data_type = NDS_DATA_TYPE_FLOAT32;
			break;
		case _64bit_double:
			data_type = NDS_DATA_TYPE_FLOAT64;
			break;
		case _32bit_complex:
			data_type = NDS_DATA_TYPE_COMPLEX32;
			break;
		default:
			data_type = 0;
	}

	return data_type;
}


static int _nds_chan_req_to_channel(nds_channel *channel, chan_req_t *chanreq)
{
	int ret = -1;

	/* Check to make sure name field will fit. */
	if (strlen(chanreq->name) >= sizeof(channel->name))
	{
		errno = NDS_CHANNEL_NAME_TOO_LONG;
		goto done;
	}

	/* Copy fields. */
	strncpy_s( channel->name,
		   sizeof( channel->name ),
		   chanreq->name,
		   sizeof(channel->name) );
	channel->channel_type = _nds_chantype_to_channel_type(chanreq->type);
	channel->data_type = _nds_data_to_data_type(chanreq->data_type);
	channel->sample_rate = chanreq->rate;

	/* Set return value to indicate success. */
	ret = 0;

done:
	/* Done! */
	return ret;
}


nds_connection *nds_open(const char *host, int port, nds_protocol protocol)
{
	/* Initialize NDS client library.  FIXME: this static initialization is not
	* entirely thread safe. */
	static int startup = 0;

	nds_connection* connection = (nds_connection*)NULL;
	char *host_copy = (char*)NULL;
	int host_copy_len = 0;
	int ret = -1;

	if (!startup)
	{
		startup = 1;
		daq_startup();
	}

	connection = malloc(sizeof(struct _nds_connection));
	host_copy_len = strlen(host) + 1;
	host_copy = malloc( host_copy_len );

	if (connection && host_copy) {
	        strcpy_s(host_copy, host_copy_len, host);
		connection->host = host_copy;
		connection->port = port;
	} else {
		free(connection);
		free(host_copy);
	}

	ret = daq_connect(&connection->daq, host, port,
			  ( protocol == NDS_PROTOCOL_TRY )
			  ? nds_try
			  : protocol );

	if (ret != 0)
	{
		free(host_copy);
		free(connection);
		connection = NULL;
		errno = NDS_ELAST + ret;
	}

	return connection;
}


void nds_close(nds_connection *connection)
{
	if (connection)
	{
		daq_disconnect(&connection->daq);
		free(connection->host);
		connection->host = NULL;
		free(connection);
	} else
		errno = EFAULT;
}


char *nds_get_host(nds_connection *connection)
{
	char *ret;

	if (connection) {
	        const int ret_len = strlen( connection->host ) + 1;
		ret = malloc( ret_len );
		if (ret)
			strcpy_s(ret, ret_len, connection->host);
	} else {
		ret = NULL;
		errno = EFAULT;
	}

	return ret;
}


int nds_get_port(nds_connection *connection)
{
	int ret;

	if (connection) {
		ret = connection->port;
	} else {
		ret = -EFAULT;
		errno = -ret;
	}

	return ret;
}


nds_protocol nds_get_protocol(nds_connection *connection)
{
	int ret;

	if (connection) {
		ret = connection->daq.nds_versn;
		if (ret == nds_try)
			ret = NDS_PROTOCOL_TRY;
	} else {
		errno = EFAULT;
		ret = -1;
	}

	return ret;
}


nds_channel **nds_find_channels(
	nds_connection *connection,
	size_t *count_channels,
	const char *channel_glob,
	nds_channel_type channel_type_mask,
	nds_data_type data_type_mask,
	double min_sample_rate,
	double max_sample_rate
)
{
	nds_channel **ret = NULL;
	int nchannels, retval;
	daq_channel_t *daq_channels = NULL;
	nds_channel **channels = NULL;
	size_t i, j = 0;

	if (!connection || !count_channels)
	{
		errno = EFAULT;
		goto fail;
	}

	retval = daq_recv_channels(&connection->daq, NULL, 0, &nchannels);

	if (retval)
	{
		errno = NDS_ELAST + retval;
		goto fail;
	}

	daq_channels = malloc(sizeof(daq_channel_t) * nchannels);

	if (!daq_channels)
		goto fail;

	retval = daq_recv_channels(&connection->daq, daq_channels,
		nchannels, &nchannels);

	if (retval)
	{
		errno = NDS_ELAST + retval;
		goto fail;
	}

	channels = malloc(sizeof(nds_channel *) * nchannels);

	if (!channels)
		goto fail;

	for (i = j = 0; i < nchannels; i ++)
	{
		daq_channel_t *daq_channel = &daq_channels[i];
		nds_channel *channel;
		nds_channel_type channel_type;
		nds_data_type data_type;

		if (daq_channel->rate < min_sample_rate)
			continue;
		if (daq_channel->rate> max_sample_rate)
			continue;

		channel_type = _nds_chantype_to_channel_type(daq_channel->type);

		if (!(channel_type & channel_type_mask))
			continue;

		data_type = _nds_data_to_data_type(daq_channel->data_type);

		if (!(data_type & data_type_mask))
			continue;

		/* FIXME: have to use Windows equivalent, PathMatchSpec */
		if (channel_glob)
		{
			retval = fnmatch(channel_glob, daq_channel->name, 0);
			if (retval == FNM_NOMATCH)
				continue;
			else if (retval != 0)
				goto fail;
		}

		channel = malloc(sizeof(nds_channel));
		if (!channel)
			goto fail;

		if (strlen(daq_channel->name) >= sizeof(channel->name))
		{
			errno = NDS_CHANNEL_NAME_TOO_LONG;
			free(channel);
			goto fail;
		}

		strncpy_s( channel->name,
			   sizeof( channel->name ),
			   daq_channel->name,
			   sizeof(channel->name) );
		channel->sample_rate = daq_channel->rate;
		channel->channel_type = channel_type;
		channel->data_type = data_type;

		channels[j ++] = channel;
	}

	ret = channels;
	channels = NULL;
	*count_channels = j;
fail:
	if (channels)
	{
		for (i = 0; i < j; i ++)
			free(channels[i]);
		free(channels);
	}
	free(daq_channels);
	return ret;
}


int nds_sizeof_data_type(nds_data_type data_type)
{
	int ret;

	switch (data_type) {
		case NDS_DATA_TYPE_INT16:
			ret = 2;
			break;
		case NDS_DATA_TYPE_INT32:
		case NDS_DATA_TYPE_FLOAT32:
			ret = 4;
			break;
		case NDS_DATA_TYPE_INT64:
		case NDS_DATA_TYPE_FLOAT64:
		case NDS_DATA_TYPE_COMPLEX32:
			ret = 8;
			break;
		default:
			ret = -EINVAL;
	}

	if (ret < 0)
		errno = -ret;
	return ret;
}


nds_buffer **nds_fetch(
	nds_connection *connection,
	long gps_start,
	long gps_stop,
	const char **channel_names,
	size_t count_channels
) {
	nds_buffer **ret = NULL;
	unsigned char **ptrs = malloc(sizeof(char *) * count_channels);
	unsigned char **ends = malloc(sizeof(char *) * count_channels);
	int result;
	size_t count_received_channels;
	size_t i;

	/* Check that all temporary arrays were allocated. */
	if (!ptrs || !ends)
		goto fail;

	/* Try to request data for indicated times and channels. */
	result = nds_iterate(connection, gps_start, gps_stop, channel_names,
		count_channels);
	if (result)
		goto fail;

	/* Try to retrieve the first set of buffers. */
	ret = nds_next(connection, &count_received_channels);
	if (!ret)
		goto fail;

	/* Check that the correct number of buffers arrived. */
	if (count_channels != count_received_channels)
	{
		errno = NDS_UNEXPECTED_CHANNELS_RECEIVED;
		goto fail;
	}

	/* Reallocate the received buffers to fit the entire request. */
	for (i = 0; i < count_channels; i ++)
	{
		nds_buffer *new_buf;
		size_t unit_size = nds_sizeof_data_type(ret[i]->channel.data_type);
		size_t needed_samples
			= (size_t)((gps_stop - gps_start) * ret[i]->channel.sample_rate);
		size_t needed_bytes = needed_samples * unit_size;
		new_buf = realloc(ret[i], sizeof(nds_buffer) + needed_bytes);
		if (!new_buf)
			goto fail;
		ret[i] = new_buf;
		ptrs[i] = ret[i]->data + ret[i]->length * unit_size;
		ends[i] = ret[i]->data + needed_bytes;
		ret[i]->length = needed_samples;
	}

	/* Retrieve more buffers as they arrive. */
	while (1) {
		chan_req_t *chan_reqs;
		char *block_data;

		/* Check to see if we need to retrieve any more data. */
		for (i = 0; i < count_channels; i ++)
			if (ptrs[i] < ends[i])
				break;
		if (i == count_channels)
			break;

		result = daq_recv_next(&connection->daq);
		if (result != DAQD_OK)
		{
			errno = NDS_ELAST + result;
			goto fail;
		}

		/* Check that the correct number of buffers arrived. */
		if (count_channels != connection->daq.num_chan_request)
		{
			errno = NDS_UNEXPECTED_CHANNELS_RECEIVED;
			goto fail;
		}

		/* Copy data buffer by buffer. */
		chan_reqs = connection->daq.chan_req_list;
		block_data = daq_get_block_data(&connection->daq);
		for (i = 0; i < count_channels; i ++)
		{
			/* Determine number of bytes received for this channel. */
			int nbytes = chan_reqs[i].status;

			/* Check that no error condition occurred on this channel. */
			if (nbytes < 0)
			{
				errno = NDS_ELAST - nbytes;
				goto fail;
			}

			/* Check that there is room to copy the new data. */
			if (ptrs[i] + nbytes > ends[i])
			{
				errno = NDS_UNEXPECTED_DATA_RECEIVED;
				goto fail;
			}

			/* Copy the new data. */
			memcpy(ptrs[i], block_data + chan_reqs[i].offset, (size_t) nbytes);
			ptrs[i] += nbytes;
		}
	}

	/* No errors; jump to end! */
	goto done;
fail:
	/* Uh-oh, something wrong happened. */
	if (ret)
	{
		for (i = 0; i < count_channels; i ++)
			free(ret[i]);
		free(ret);
		ret = NULL;
	}
done:
	/* Done! */
	free(ptrs);
	free(ends);
	return ret;
}


int nds_iterate(
	nds_connection *connection,
	long gps_start,
	long gps_stop,
	const char **channel_names,
	size_t count_channels
)
{
	int ret = -1;
	int result;
	size_t i;

	if (connection)
	{
		daq_clear_channel_list(&connection->daq);

		switch (nds_get_protocol(connection))
		{
			case NDS_PROTOCOL_ONE:
				{
					int nchannels;
					daq_channel_t *daq_channels = NULL;

					result = daq_recv_channels(&connection->daq, NULL, 0,
						&nchannels);

					if (result != DAQD_OK)
					{
						errno = NDS_ELAST + result;
						break;
					}

					daq_channels = malloc(sizeof(daq_channel_t) * nchannels);

					if (!daq_channels)
						break;

					result = daq_recv_channels(&connection->daq,
						daq_channels, nchannels, &nchannels);

					if (result != DAQD_OK)
					{
						free(daq_channels);
						break;
					}

					for (i = 0; i < count_channels; i ++)
					{
						int j;
						int channel_found = 0;
						for (j = 0; j < nchannels; j ++)
						{
							if (strcmp(channel_names[i],
									   daq_channels[j].name) == 0)
							{
								result = daq_request_channel_from_chanlist(
									&connection->daq, &daq_channels[j]);
								if (result == DAQD_OK)
									channel_found = 1;
								else
									errno = NDS_ELAST + result;
								break;
							}
						}
						if (!channel_found)
						{
							errno = NDS_ELAST + DAQD_INVALID_CHANNEL_NAME;
							break;
						}
					}

					free(daq_channels);
					if (i == count_channels)
						ret = 0;
				}
				break;
			case NDS_PROTOCOL_TWO:
				{
					for (i = 0, result = DAQD_OK; i < count_channels; i ++)
					{
						result = daq_request_channel(&connection->daq,
							channel_names[i], cUnknown, 0.);
					}
					if (result == DAQD_OK)
						ret = 0;
					else
						errno = NDS_ELAST + result;
				}
				break;
			case -1:
				break;
			case NDS_PROTOCOL_TRY:
			default:
				errno = NDS_ELAST + DAQD_NOT_CONFIGURED;
				break;
		}
	} else
		errno = EFAULT;

	if (ret == 0)
	{
		result = daq_request_data(&connection->daq, (time_t) gps_start, (time_t) gps_stop, (time_t) 1);
		if (result != DAQD_OK)
		{
			errno = NDS_ELAST + result;
			ret = -1;
		}
	}

	return ret;
}


nds_buffer **nds_next(
	nds_connection *connection,
	size_t *count_channels
)
{
	nds_buffer **ret = NULL;
	size_t count_received_channels;
	chan_req_t *chan_reqs;
	size_t i;
	int result;

	if (!connection)
	{
		errno = EFAULT;
		goto done;
	}

	result = daq_recv_next(&connection->daq);
	if (result != DAQD_OK)
	{
		errno = NDS_ELAST + result;
		goto done;
	}

	count_received_channels = connection->daq.num_chan_request;

	ret = malloc(sizeof(nds_buffer *) * count_received_channels);
	if (!ret)
		goto done;

	chan_reqs = connection->daq.chan_req_list;

	for (i = 0; i < count_received_channels; i ++)
	{
		int nbytes = chan_reqs[i].status;

		if (nbytes < 0)
		{
			errno = NDS_ELAST - nbytes;
			break;
		}

		/* Allocate buffer, making room for data payload. */
		ret[i] = malloc(sizeof(nds_buffer) + nbytes);
		if (!ret[i])
			break;

		/* Copy channel information. */
		result = _nds_chan_req_to_channel(&ret[i]->channel, &chan_reqs[i]);
		if (result)
		{
			free(ret[i]);
			break;
		}

		/* Set buffer's metadata. */
		ret[i]->gps_seconds = daq_get_block_gps(&connection->daq);
		ret[i]->gps_nanoseconds = daq_get_block_gpsn(&connection->daq);
		ret[i]->length = nbytes / nds_sizeof_data_type(ret[i]->channel.data_type);

		/* Copy buffer payload. */
		memcpy(ret[i]->data, daq_get_block_data(&connection->daq) + chan_reqs[i].offset, (size_t) nbytes);
	}

	if (i == count_received_channels)
	{
		*count_channels = count_received_channels;
	} else {
		size_t j;
		for (j = 0; j < i; j ++)
			free(ret[j]);
		free(ret);
		ret = NULL;
	}

done:
	return ret;
}


void nds_expire_cache(nds_connection *connection)
{
	errno = ENOSYS;
}


void nds_perror(const char *s)
{
	fprintf(stderr, "%s: %s\n", s, nds_strerror(errno));
}


const char *nds_strerror(int errnum)
{
	const char *errstr;

	switch (errnum) {
		case NDS_EOS:
			errstr = "Reached end of stream.";
			break;
		case NDS_CHANNEL_NAME_TOO_LONG:
			errstr = "Channel name is too long.";
			break;
		case NDS_UNEXPECTED_CHANNELS_RECEIVED:
			errstr = "Server sent more channels than were expected.";
		case NDS_UNEXPECTED_DATA_RECEIVED:
			errstr = "Server sent more data than were expected.";
		default:
			if (errnum >= NDS_ELAST)
				errstr = daq_strerror(errnum - NDS_ELAST);
			else
				errstr = strerror(errnum);
	}

	return errstr;
}

#if FNMATCH_STATIC
static int fnmatch( const char* Pattern,
		    const char* Name,
		    int Flags )
{
  /* Written by Jack Handy - <A href="mailto:jakkhandy@hotmail.com">jakkhandy@hotmail.com</A> */
  /* http://www.codeproject.com/Articles/1088/Wildcard-string-compare-globbing */

  /*
   * This code is provided for Windows system as they lack an fnmatch() function.
   * It is used instead of PathMatchSpec() so as not to require C++ compilation or linking.
   */
  const char *cp = NULL, *mp = NULL;

  while ((*Name) && (*Pattern != '*')) {
    if ((*Pattern != *Name) && (*Pattern != '?')) {
      return 0;
    }
    Pattern++;
    Name++;
  }

  while (*Name) {
    if (*Pattern == '*') {
      if (!*++Pattern) {
        return 1;
      }
      mp = Pattern;
      cp = Name+1;
    } else if ((*Pattern == *Name) || (*Pattern == '?')) {
      Pattern++;
      Name++;
    } else {
      Pattern = mp;
      Name = cp++;
    }
  }

  while (*Pattern == '*') {
    Pattern++;
  }
  return ( ( *Pattern == '\0' )
	   ? 0
	   : FNM_NOMATCH );
}
#endif /* FNMATCH_STATIC */
